#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "assembler.h"
#include "codegen.h"
#include "dlllinker.h"
#include "errorreporter.h"
#include "optimizer.h"
#include "parser.h"
#include "stringlinker.h"
#include "symtab.h"
#include "tokenizer.h"
#include "warningreporter.h"

extern _token tokens;
extern _codelist codelist;
extern _symtab symtab[6];

extern bool bEOF;
extern long *plnLongsInFunction;

extern long *lpErrors;
extern long lnErrors;
extern bool bError;

extern long lnFunctions;

#include "consola.h"

long lOptimize = 3;
long lOptimizeType = 3;
bool bWarnings = true;
bool bWarningsAsErrors = false;
long lErrorCorrection = 0;
bool bPreserveSymbols = false;
bool bLog = false;
char szOutputFile[0x100];
bool bSpecifiedOutput = false;

extern _dlllink		*LinkedDLLProcs;
extern long			lnDLLLinks;
extern char			*pszDLLLinkData;

//extern bool			bDLLLinkData;
extern long			lSizeDLLLinkData;

extern _stringlink	*LinkedStrings;
extern long			lnStringLinks;
extern char			*pszStringLinkData;

//extern bool			bStringLinkData;
extern long			lSizeStringLinkData;

time_t tTime;

extern long	lnWarnings;
extern long *lpWarnings;

int main(int argc, char *argv[])
	{
	tTime = time(&tTime);

	printf("\n BASM v 0.8.1.1\n-==--==--==--==-\n");

	printf("Start: %s", ctime(&tTime));

	if (argc < 2 || argc > 8)
		{
		printf("Usage: basm <basmfile> [cobfile] [flags]\n");
		printf("cobfile: default is same as input file\n");
		printf("flags:\t-oN: set optimization level to N\t-tN: set optimization type to N\n");
		printf("\t-w:  don't display warnings\t\t-e:  treat warnings as errors\n");
		printf("\t-cN: set error correction level to N\t-l:  turn logging on\n");
		printf("\t-s:  preserve symbol data\n");
		printf("Some options are not available yet. See BASM documentation for more information.\n");
		exit(0);
		}

	FILE *fp = fopen(argv[1], "rb");

	if (!fp)
		{
		printf("Bad file name: %s\n", argv[1]);
		exit(0);
		}

	strcpy(szOutputFile, argv[1]);
	long lFileLen = strlen(szOutputFile);

	if (lFileLen < 4)
		lFileLen = 4;

	szOutputFile[lFileLen] = NULL;
	szOutputFile[lFileLen - 1] = NULL;
	szOutputFile[lFileLen - 2] = 'b';
	szOutputFile[lFileLen - 3] = 'o';
	szOutputFile[lFileLen - 4] = 'c';
	szOutputFile[lFileLen - 5] = '.';

	//get the commandline
	for (long i = 2; i < argc; i++)
		{
		if (argv[i][0] == '-') //found a flag
			{
			switch(argv[i][1])
				{
				case 'o': //optimization level
					lOptimize = argv[i][2];
					break;
				case 't': //optimization type
					lOptimizeType = argv[i][2];
					break;
				case 'w': //warnings off
					bWarnings = 0;
					break;
				case 'e': //warnings as errors
					bWarningsAsErrors = 1;
					break;
				case 'c': //error correction level
					lErrorCorrection = argv[i][2];
					break;
				case 's': //preserve symbol data
					bPreserveSymbols = true;
					break;
				case 'l': //turn logging on
					bLog = true;
					break;
				default:
					printf("Unrecognized switch, -%c\n", argv[i][1]);
					break;
				}
			}
		else //found a file name
			{
			bSpecifiedOutput = true;
			strcpy(szOutputFile, argv[i]);
			}
		}

	if (bLog)
		printf("BASM will log errors to \"errors.txt\"\n");

	printf("BASM will output to \"%s\"\n\n", szOutputFile);

	printf("File: %s\n", argv[1]);
//	printf("Preprocessing . . .\n");
//	preprocess(fp);
	printf("Tokenizing . . .\n");
	tokenize(fp);
	printf("Parsing . . .\n");
	BASMFile(&tokens);

	if (lnErrors)
		{
		textcolor(LIGHTRED);
		printf("%d errors found!\n", lnErrors);
		textcolor(LIGHTGRAY);
		//if (lnErrors)
		reporterrors(argv[1]);
		}
	else
		{
		if (lnWarnings)
			{
			textcolor(YELLOW);
			printf("%d warnings found!\n", lnWarnings);
			textcolor(LIGHTGRAY);
			reportwarnings(argv[1]);

			if (bWarningsAsErrors)
				goto SkipCompile;
			}

//		bDLLLinkData = false;
		lSizeDLLLinkData = 0;
		printf("No errors found\nGenerating DLL Link Table . . .\n");
		CreateLinkTable();

		lSizeStringLinkData = 0;
		printf("Generating String Link Table . . .\n");
		CreateStringLinkTable();

		if (lOptimize && lOptimizeType)
			{
			printf("Optimizing (1st Pass - Pre Code Generation) - Level = %d, Type = %d . . .\n", lOptimize, lOptimizeType);
			optimize_pre(lOptimize, lOptimizeType);
			}

		printf("Generating code . . .\n");
		generatecode();

		if (lOptimize && lOptimizeType)
			{
			printf("Optimizing (2nd Pass - Post Code Generation) - Level = %d, Type = %d . . .\n", lOptimize, lOptimizeType);
			optimize_post(lOptimize, lOptimizeType);
			}

		printf("Assembling . . .\n");
		assemblecob(szOutputFile);
		printf("Done!\n");
		}

SkipCompile:

	fclose(fp);

	_token *pTok = &tokens;
	_token *pNextTok;
	while (pTok->pNext)
		{
		if (pTok->pNext->psz)
			free(pTok->pNext->psz);

		pNextTok = pTok->pNext->pNext;
		free(pTok->pNext);
		pTok->pNext = pNextTok;
		}
	
	if (pTok->psz) free(pTok->psz);

	_codelist *pCodelist = &codelist;
	_codelist *pNextCodelist;
	while (pCodelist->pNext)
		{
		if (pCodelist->pNext)
			{
			pNextCodelist = pCodelist->pNext->pNext;
			free(pCodelist->pNext);
			pCodelist->pNext = pNextCodelist;
			}
		}

	_symtab *pSymtab;
	_symtab *pNextSymtab;
	for (long j = 0; j < 7; j++)
		{
		pSymtab = &symtab[j];

		while (pSymtab->pNext)
			{
			if (pSymtab->pNext->tok.psz)
				free(pSymtab->pNext->tok.psz);

			pNextSymtab = pSymtab->pNext->pNext;
			free(pSymtab->pNext);
			pSymtab->pNext = pNextSymtab;
			}

		if (pSymtab->tok.psz) free(pSymtab->tok.psz);
		}

	if (lnFunctions)
		delete[] plnLongsInFunction;

	if (lnErrors)
		free(lpErrors);

	for (long k = 0; k < lnDLLLinks; k++)
		{
		if (LinkedDLLProcs[k].psz)
			delete[] LinkedDLLProcs[k].psz;
		}

	if (LinkedDLLProcs)
		free(LinkedDLLProcs);

	if (pszDLLLinkData)
		delete[] pszDLLLinkData;

	for (long l = 0; l < lnStringLinks; l++)
		{
		if (LinkedStrings[l].psz)
			delete[] LinkedStrings[l].psz;
		}

	if (LinkedStrings)
		free(LinkedStrings);

	if (pszStringLinkData)
		delete[] pszStringLinkData;

	tTime = time(&tTime);
	printf("Finish: %s", ctime(&tTime));

	return 0;
	}